package edu.uci.eecs.specCompiler.codeGenerator;

import java.util.ArrayList;
import java.util.HashSet;
import java.io.File;

import edu.uci.eecs.specCompiler.specExtraction.ConditionalInterface;
import edu.uci.eecs.specCompiler.specExtraction.GlobalConstruct;
import edu.uci.eecs.specCompiler.specExtraction.IDExtractor;
import edu.uci.eecs.specCompiler.specExtraction.InterfaceConstruct;
import edu.uci.eecs.specCompiler.specExtraction.PotentialCPDefineConstruct;
import edu.uci.eecs.specCompiler.specExtraction.SequentialDefineSubConstruct;
import edu.uci.eecs.specCompiler.specExtraction.SpecExtractor;
import edu.uci.eecs.specCompiler.specExtraction.VariableDeclaration;

public class CodeVariables {
	// C++ code or library
	public static final String HEADER_THREADS = "<threads.h>";
	public static final String HEADER_STDINT = "<stdint.h>";
	public static final String ThreadIDType = "thrd_t";
	public static final String GET_THREAD_ID = "thrd_current";
	public static final String BOOLEAN = "bool";
	public static final String UINT64 = "uint64_t";

	// Model checker code
	public static final String HEADER_CDSANNOTATE = "<cdsannotate.h>";
	public static final String HEADER_SPECANNOTATION = "<specannotation.h>";
	public static final String HEADER_CDSTRACE = "<cdstrace.h>";
	public static final String CDSAnnotate = "cdsannotate";
	public static final String CDSAnnotateType = "SPEC_ANALYSIS";

	public static final String SPEC_ANNO_TYPE = "spec_anno_type";
	public static final String SPEC_ANNO_TYPE_HB_INIT = "HB_INIT";
	public static final String SPEC_ANNO_TYPE_INTERFACE_BEGIN = "INTERFACE_BEGIN";
	public static final String SPEC_ANNO_TYPE_HB_CONDITION = "HB_CONDITION";
	public static final String SPEC_ANNO_TYPE_INTERFACE_END = "INTERFACE_END";
	public static final String SPEC_ANNO_TYPE_POTENTIAL_CP_DEFINE = "POTENTIAL_CP_DEFINE";
	public static final String SPEC_ANNOTATION = "spec_annotation";
	public static final String SPEC_ANNOTATION_FIELD_TYPE = "type";
	public static final String SPEC_ANNOTATION_FIELD_ANNO = "annotation";

	public static final String ANNO_HB_INIT = "anno_hb_init";
	public static final String ANNO_INTERFACE_BOUNDARY = "anno_interface_boundary";
	public static final String ANNO_ID = "anno_id";
	public static final String ANNO_POTENTIAL_CP_DEFINE = "anno_potentail_cp_define";
	public static final String ANNO_CP_DEFINE = "anno_cp_define";
	public static final String ANNO_CP_DEFINE_CHECK = "anno_cp_define_check";
	public static final String ANNO_HB_CONDITION = "anno_hb_condition";
	public static final String ANNO_POST_CHECK = "anno_post_check";

	// Specification variables
	public static final String SPEC_INTERFACE_WRAPPER = "__wrapper_";

	// Specification library
	public static final String HEADER_SPEC_LIB = "<spec_lib.h>";
	public static final String SPEC_QUEUE = "spec_queue";
	public static final String SPEC_STACK = "spec_stack";
	public static final String SPEC_DEQUE = "spec_deque";
	public static final String SPEC_HASHTABLE = "spec_hashtable";
	public static final String SPEC_PRIVATE_HASHTABLE = "spec_private_hashtable";
	public static final String SPEC_TAG = "spec_tag";
	public static final String SPEC_TAG_CURRENT = "current";
	public static final String SPEC_TAG_NEXT = "next";

	// Macro
	public static final String MACRO_ID = "__ID__";
	public static final String MACRO_COND = "__COND_SAT__";
	public static final String MACRO_RETURN = "__RET__";
	public static final String MACRO_ATOMIC_RETURN = "__ATOMIC_RET__";

	public static void printCode(ArrayList<String> code) {
		for (int i = 0; i < code.size(); i++) {
			System.out.println(code.get(i));
		}
	}

	private static String COMMENT(String comment) {
		return "/* " + comment + " */";
	}

	private static String INCLUDE(String header) {
		return "#include " + header;
	}

	private static String DEFINE(String left, String right) {
		return "#define " + left + " " + right;
	}

	private static String UNDEFINE(String macro) {
		return "#undef " + macro;
	}

	private static String BRACE(String val) {
		return "(" + val + ")";
	}

	private static String ASSIGN(String structName, String field, String val) {
		return structName + "." + field + " = " + val + ";";
	}

	private static String ASSIGN_PTR(String structName, String field, String val) {
		return structName + "." + field + " = &" + val + ";";
	}

	private static String DECLARE(String type, String name) {
		return type + " " + name + ";";
	}

	private static String DECLARE_DEFINE(String type, String var, String val) {
		return type + " " + var + " = " + val + ";";
	}

	private static String ANNOTATE(String structName) {
		return CDSAnnotate + "(" + CDSAnnotateType + ", &" + structName + ");";
	}

	private static HashSet<String> getAllHeaders(SemanticsChecker semantics) {
		HashSet<String> headers = new HashSet<String>();
		for (String interfaceName : semantics.interfaceName2Construct.keySet()) {
			File f = semantics.interfaceName2Construct.get(interfaceName).file;
			headers.addAll(semantics.srcFilesInfo.get(f).headers);
		}
		return headers;
	}

	public static ArrayList<String> generateGlobalVarDeclaration(
			SemanticsChecker semantics, GlobalConstruct construct) {
		ArrayList<String> newCode = new ArrayList<String>();
		HashSet<String> allHeaders = getAllHeaders(semantics);

		// All headers needed by the interface decalration
		newCode.add(COMMENT("/* Include all the header files that contains the interface declaration */"));
		for (String header : allHeaders) {
			newCode.add(INCLUDE(header));
		}
		newCode.add("");
		// Other necessary headers
		newCode.add(INCLUDE(HEADER_STDINT));
		newCode.add(INCLUDE(HEADER_CDSANNOTATE));
		newCode.add(INCLUDE(HEADER_SPEC_LIB));
		newCode.add("");

		SequentialDefineSubConstruct code = construct.code;
		// User-defined functions
		newCode.add(COMMENT("All other user-defined functions"));
		ArrayList<ArrayList<String>> defineFuncs = code.defineFuncs;
		for (int i = 0; i < defineFuncs.size(); i++) {
			newCode.addAll(defineFuncs.get(i));
			newCode.add("");
		}

		// Define necessary info structure
		newCode.add(COMMENT("Definition of interface info struct (per interface)"));
		

		// User-defined variables
		ArrayList<VariableDeclaration> varDecls = code.declareVar;
		for (int i = 0; i < varDecls.size(); i++) {
			VariableDeclaration varDecl = varDecls.get(i);
			newCode.add(DECLARE(varDecl.type, varDecl.name));
		}

		// printCode(newCode);
		return newCode;
	}

	private static ArrayList<String> generateHBInitAnnotation(
			SemanticsChecker semantics) {
		ArrayList<String> newCode = new ArrayList<String>();
		int hbConditionInitIdx = 0;
		for (ConditionalInterface left : semantics.getHBConditions().keySet()) {
			for (ConditionalInterface right : semantics.getHBConditions().get(
					left)) {
				String structVarName = "hbConditionInit" + hbConditionInitIdx;
				String annotationVarName = "hb_init" + hbConditionInitIdx;
				hbConditionInitIdx++;
				String interfaceNumBefore = Integer
						.toString(semantics.interface2Num
								.get(left.interfaceName)), hbLabelNumBefore = Integer
						.toString(semantics.hbLabel2Num
								.get(left.hbConditionLabel)), interfaceNumAfter = Integer
						.toString(semantics.interface2Num
								.get(right.interfaceName)), hbLabelNumAfter = Integer
						.toString(semantics.hbLabel2Num
								.get(right.hbConditionLabel));
				newCode.add(COMMENT(left + " -> " + right));

				newCode.add(ANNO_HB_INIT + " " + structVarName + ";");
				newCode.add(ASSIGN(structVarName, "interface_num_before",
						interfaceNumBefore));
				newCode.add(ASSIGN(structVarName, "hb_condition_num_before",
						hbLabelNumBefore));
				newCode.add(ASSIGN(structVarName, "interface_num_after",
						interfaceNumAfter));
				newCode.add(ASSIGN(structVarName, "hb_condition_num_after",
						hbLabelNumAfter));

				newCode.add(DECLARE(SPEC_ANNOTATION, annotationVarName));
				newCode.add(ASSIGN(annotationVarName,
						SPEC_ANNOTATION_FIELD_TYPE, SPEC_ANNO_TYPE_HB_INIT));
				newCode.add(ASSIGN_PTR(annotationVarName,
						SPEC_ANNOTATION_FIELD_ANNO, structVarName));
				newCode.add(ANNOTATE(annotationVarName));
			}
		}
		return newCode;
	}

	public static ArrayList<String> generateInterfaceWrapper(
			SemanticsChecker semantics, InterfaceConstruct construct) {
		ArrayList<String> newCode = new ArrayList<String>();

		// Generate necessary header file (might be redundant but never mind)

		// Generate wrapper header

		// Wrapper function body

		// Interface begin

		// Call original renamed function

		// HB conditions

		// Interface end

		// End of the wrapper function

		// printCode(newCode);
		return newCode;
	}

	public static ArrayList<String> generatePotentialCPDefine(
			SemanticsChecker semantics, PotentialCPDefineConstruct construct) {
		ArrayList<String> newCode = new ArrayList<String>();

		// Generate redundant header files
		newCode.add(COMMENT("Automatically generated code for potential commit point: "
				+ construct.label));
		newCode.add(COMMENT("Include redundant headers"));
		newCode.add(INCLUDE(HEADER_THREADS));
		newCode.add(INCLUDE(HEADER_CDSTRACE));
		newCode.add("");
		// Some necessary function calls

		return newCode;
	}
}
